import grpc
from .futures import InstantFuture


class _RequestWrapper(grpc.UnaryUnaryMultiCallable):
    def __init__(self, func):
        self.func = func

    def __call__(self, request, *args, **kwargs):
        return self.func(request, *args, **kwargs)

    def with_call(self, request, *args, **kwargs):
        return self(request, *args, **kwargs)

    def future(self, request, *args, **kwargs):
        resp = self(request, *args, **kwargs)
        return InstantFuture(resp)



import pytest
from google.cloud import ndb
from google.cloud.datastore_v1.proto import datastore_pb2_grpc
from google.cloud.ndb import _datastore_api
from _pytest.monkeypatch import MonkeyPatch

from InMemoryCloudDatastoreStub import datastore_stub


@pytest.fixture(autouse=True)
def init_ndb_env_vars(monkeypatch: MonkeyPatch) -> None:
    """
    Initializing an ndb Client in a test env requires some environment variables to be set
    For now, these are just garbage values intended to give the library _something_
    (we don't expect them to actually work yet)
    """

    monkeypatch.setenv("DATASTORE_EMULATOR_HOST", "localhost")
    monkeypatch.setenv("DATASTORE_DATASET", "datastore-stub-test")


@pytest.fixture(autouse=True)
def ndb_stub(monkeypatch: MonkeyPatch) -> datastore_stub.LocalDatastoreStub:
    stub = datastore_stub.LocalDatastoreStub()

    def mock_stub() -> datastore_pb2_grpc.DatastoreStub:
        return stub

    monkeypatch.setattr(_datastore_api, "stub", mock_stub)
    return stub


@pytest.fixture(autouse=True)
def ndb_context(init_ndb_env_vars):
    client = ndb.Client()
    with client.context() as context:
        yield context



from InMemoryCloudDatastoreStub import datastore_stub
from tests.models import SimpleModel


def test_get_or_insert_existing_by_id(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model = SimpleModel(id="test", str_prop="asdf",)
    ndb_stub._insert_model(model)

    model_res = SimpleModel.get_or_insert("test")
    assert model_res == model


def test_get_or_insert_doesnt_exist() -> None:
    model_res = SimpleModel.get_or_insert("test")
    assert model_res is not None


def test_get_or_insert_then_update() -> None:
    model = SimpleModel.get_or_insert("test", int_prop=10)
    assert model is not None

    model.int_prop = 20
    model.put()

    sanity_check = SimpleModel.get_by_id("test")
    assert sanity_check == model



import pytest
from InMemoryCloudDatastoreStub import datastore_stub
from google.cloud import ndb
from tests.models import SimpleModel


def test_get_nonexistent_key() -> None:
    model = SimpleModel.get_by_id("notfound")
    assert model is None


def test_get_existing_by_id(ndb_stub: datastore_stub.LocalDatastoreStub) -> None:
    model = SimpleModel(id="test", str_prop="asdf",)
    ndb_stub._insert_model(model)

    model_res = SimpleModel.get_by_id("test")
    assert model_res == model


def test_get_existing_by_field(ndb_stub: datastore_stub.LocalDatastoreStub) -> None:
    model = SimpleModel(id="test", str_prop="asdf",)
    ndb_stub._insert_model(model)

    query_res = SimpleModel.query(SimpleModel.str_prop == "asdf").get()
    assert query_res == model


def test_get_existing_by_multi_field(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model = SimpleModel(id="test", str_prop="asdf", int_prop=42)
    ndb_stub._insert_model(model)

    query_res = SimpleModel.query(
        SimpleModel.str_prop == "asdf", SimpleModel.int_prop == 42
    ).get()
    assert query_res == model


def test_get_existing_by_field_not_found(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model = SimpleModel(id="test", str_prop="asdf",)
    ndb_stub._insert_model(model)

    query_res = SimpleModel.query(SimpleModel.str_prop == "foo").get()
    assert query_res is None


def test_fetch_existing_by_field_multiple(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model1 = SimpleModel(id="test", str_prop="asdf",)
    model2 = SimpleModel(id="test2", str_prop="asdf",)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    # We don't pass an ordering here, so the order is not deterministic
    query_res = SimpleModel.query(SimpleModel.str_prop == "asdf").fetch(limit=2)
    assert len(query_res) == 2
    query_res.sort(key=lambda m: m.key.id())
    assert query_res == [model1, model2]


def test_fetch_existing_by_field_multiple_with_order(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model1 = SimpleModel(id="test", str_prop="asdf", int_prop=10)
    model2 = SimpleModel(id="test2", str_prop="asdf", int_prop=20)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    # We don't pass an ordering here, so the order is not deterministic
    query_res = (
        SimpleModel.query(SimpleModel.str_prop == "asdf")
        .order(SimpleModel.int_prop)
        .fetch(limit=2)
    )
    assert len(query_res) == 2
    assert query_res == [model1, model2]


def test_fetch_existing_by_field_with_limit(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model1 = SimpleModel(id="test", str_prop="asdf",)
    model2 = SimpleModel(id="test2", str_prop="asdf",)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    # Since we don't pass an order by, it's non deterministic which we get
    query_res = SimpleModel.query(SimpleModel.str_prop == "asdf").fetch(limit=1)
    assert len(query_res) == 1
    assert query_res[0].str_prop == "asdf"


def test_fetch_existing_by_field_with_limit_not_hit(
    ndb_stub: datastore_stub.LocalDatastoreStub,
) -> None:
    model1 = SimpleModel(id="test", str_prop="asdf",)
    model2 = SimpleModel(id="test2", str_prop="asdfz",)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    # Since we don't pass an order by, it's non deterministic which we get
    query_res = SimpleModel.query(SimpleModel.str_prop == "asdf").fetch(limit=2)
    assert len(query_res) == 1
    assert query_res[0].str_prop == "asdf"


def test_fetch_existing_by_gt(ndb_stub: datastore_stub.LocalDatastoreStub) -> None:
    model1 = SimpleModel(id="test", int_prop=42,)
    model2 = SimpleModel(id="test2", int_prop=43,)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    query_res = SimpleModel.query(SimpleModel.int_prop > 42).fetch(limit=5)
    assert query_res == [model2]


def test_fetch_existing_by_ge(ndb_stub: datastore_stub.LocalDatastoreStub) -> None:
    model1 = SimpleModel(id="test", int_prop=42,)
    model2 = SimpleModel(id="test2", int_prop=43,)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    query_res = SimpleModel.query(SimpleModel.int_prop >= 42).fetch(limit=5)
    assert len(query_res) == 2


def test_fetch_existing_by_lt(ndb_stub: datastore_stub.LocalDatastoreStub) -> None:
    model1 = SimpleModel(id="test", int_prop=42,)
    model2 = SimpleModel(id="test2", int_prop=43,)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    query_res = SimpleModel.query(SimpleModel.int_prop < 43).fetch(limit=5)
    assert query_res == [model1]


def test_fetch_existing_by_le(ndb_stub: datastore_stub.LocalDatastoreStub) -> None:
    model1 = SimpleModel(id="test", int_prop=42,)
    model2 = SimpleModel(id="test2", int_prop=43,)
    ndb_stub._insert_model(model1)
    ndb_stub._insert_model(model2)

    query_res = SimpleModel.query(SimpleModel.int_prop <= 43).fetch(limit=5)
    assert len(query_res) == 2


def test_put_model() -> None:
    model = SimpleModel(id="test", str_prop="asdf")
    key = model.put()

    assert key == model.key


def test_put_model_matches_key_get() -> None:
    model = SimpleModel(id="test", str_prop="asdf")
    key = model.put()

    get_resp = key.get()
    assert get_resp == model


def test_put_model_matches_point_query() -> None:
    model = SimpleModel(id="test", str_prop="asdf")
    key = model.put()

    get_resp = SimpleModel.get_by_id("test")
    assert get_resp == model
    assert model.key == key


def test_query_all() -> None:
    stored_keys = ndb.put_multi(
        [SimpleModel(id=f"test{i}", int_prop=i) for i in range(10)]
    )
    assert len(stored_keys) == 10

    query_all = SimpleModel.query().order(SimpleModel.int_prop).fetch()
    assert len(query_all) == 10
    for i, model in enumerate(query_all):
        assert model.int_prop == i


def test_query_in_range() -> None:
    stored_keys = ndb.put_multi(
        [SimpleModel(id=f"test{i}", int_prop=i) for i in range(10)]
    )
    assert len(stored_keys) == 10

    resp = (
        SimpleModel.query(SimpleModel.int_prop >= 1, SimpleModel.int_prop < 5)
        .order(SimpleModel.int_prop)
        .fetch()
    )
    assert len(resp) == 4
    for i, model in enumerate(resp):
        assert model.int_prop == i + 1


def test_query_keys_only() -> None:
    stored_keys = ndb.put_multi(
        [SimpleModel(id=f"test{i}", int_prop=i) for i in range(10)]
    )
    assert len(stored_keys) == 10

    resp = SimpleModel.query().fetch(keys_only=True)
    assert len(resp) == 10
    assert all(isinstance(i, ndb.Key) for i in resp)


def test_query_projection() -> None:
    stored_keys = ndb.put_multi(
        [SimpleModel(id=f"test{i}", int_prop=i, str_prop=f"{i}") for i in range(10)]
    )
    assert len(stored_keys) == 10

    resp = SimpleModel.query().fetch(projection=[SimpleModel.int_prop])
    assert len(resp) == 10
    assert all(i.int_prop is not None for i in resp)
    with pytest.raises(ndb.model.UnprojectedPropertyError):
        assert all(i.str_prop is None for i in resp)



import google.cloud.datastore.helpers as ds_helpers
from google.cloud import ndb
from google.cloud.datastore_v1 import types
from google.cloud.datastore_v1.proto import datastore_pb2_grpc
from typing import List

from ._in_memory_store import _InMemoryStore
from ._request_wrapper import _RequestWrapper
from ._stored_object import _StoredObject
from ._transactions import _TransactionType


class LocalDatastoreStub(datastore_pb2_grpc.DatastoreStub):

    _OPERATOR_TO_CMP_METHOD_NAME = {
        types.PropertyFilter.Operator.LESS_THAN: "__lt__",
        types.PropertyFilter.Operator.LESS_THAN_OR_EQUAL: "__le__",
        types.PropertyFilter.Operator.GREATER_THAN: "__gt__",
        types.PropertyFilter.Operator.GREATER_THAN_OR_EQUAL: "__ge__",
        types.PropertyFilter.Operator.EQUAL: "__eq__",
        # TODO missing HAS_ANCESTOR
    }

    store: _InMemoryStore

    Lookup: _RequestWrapper
    Commit: _RequestWrapper
    RunQuery: _RequestWrapper
    BeginTransaction: _RequestWrapper
    Rollback: _RequestWrapper
    # AllocateIds: _RequestWrapper
    # ReserveIds: _RequestWrapper

    def __init__(self) -> None:
        self.store = _InMemoryStore()

        self.Lookup = _RequestWrapper(self._lookup)
        self.Commit = _RequestWrapper(self._commit)
        self.RunQuery = _RequestWrapper(self._run_query)
        self.BeginTransaction = _RequestWrapper(self._begin_transaction)
        self.Rollback = _RequestWrapper(self._rollback)

    def _insert_model(self, model: ndb.Model) -> None:
        ds_key = model.key._key.to_protobuf()
        assert self.store.get(ds_key, None) is None
        entity_proto = ndb.model._entity_to_protobuf(model)
        self.store.put(entity_proto, 0, None)

    def _lookup(
        self, request: types.LookupRequest, *args, **kwargs
    ) -> types.LookupResponse:
        found: List[types.EntityResult] = []
        missing: List[types.EntityResult] = []
        transaction_id = request.read_options.transaction

        for key in request.keys:
            stored_data = self.store.get(key, transaction_id)
            if stored_data:
                found.append(
                    types.EntityResult(
                        entity=stored_data.entity, version=stored_data.version
                    )
                )
            else:
                missing.append(
                    types.EntityResult(
                        entity=types.Entity(key=key),
                        version=self.store.seqid(transaction_id),
                    )
                )

        return types.LookupResponse(found=found, missing=missing,)

    def _begin_transaction(
        self, request: types.BeginTransactionRequest, *args, **kwargs
    ) -> types.BeginTransactionResponse:
        transaction_mode = None
        request_type = request.transaction_options.WhichOneof("mode")
        if request_type == "read_write":
            transaction_mode = _TransactionType.READ_WRITE
        elif request_type == "read_only":
            transaction_mode = _TransactionType.READ_ONLY
        assert transaction_mode is not None
        transaction_id = self.store.beginTransaction(transaction_mode)

        return types.BeginTransactionResponse(transaction=transaction_id,)

    def _commit(
        self, request: types.CommitRequest, *args, **kwargs
    ) -> types.CommitResponse:
        results: List[types.MutationResult] = self.store.commitTransaction(
            request.transaction, request.mutations
        )
        return types.CommitResponse(mutation_results=results, index_updates=0,)

    def _rollback(
        self, request: types.RollbackRequest, *args, **kwargs
    ) -> types.RollbackResponse:
        self.store.rollbackTransaction(request.transaction)
        return types.RollbackResponse()

    def _run_query(
        self, request: types.RunQueryRequest, *args, **kwargs
    ) -> types.RunQueryResponse:
        # Don't support cloud sql
        # TODO also figire out error handling
        assert request.query

        # Query processing will be very naive.
        query: types.Query = request.query
        transaction_id: bytes = request.read_options.transaction
        resp_data: List[_StoredObject] = []
        for _, stored in self.store.items(transaction_id):
            if query.kind and stored.entity.key.path[-1].kind != query.kind[0].name:
                # this doesn't account for ancestor keys
                continue

            if self._matches_filter(stored, query.filter):
                resp_data.append(stored)

        if query.order:
            # TODO
            assert len(query.order) == 1
            order = query.order[0]
            assert order.direction in [
                types.PropertyOrder.Direction.DESCENDING,
                types.PropertyOrder.Direction.ASCENDING,
            ]
            resp_data.sort(
                key=lambda d: ds_helpers._get_value_from_value_pb(
                    d.entity.properties.get(order.property.name)
                ),
                reverse=order.direction == types.PropertyOrder.Direction.DESCENDING,
            )

        if query.HasField("limit"):
            resp_data = resp_data[: query.limit.value]

        if query.projection:
            projection_fields = [p.property.name for p in query.projection]
            if projection_fields == ["__key__"]:
                result_type = types.EntityResult.ResultType.KEY_ONLY
                entity_results = [
                    types.EntityResult(
                        entity=types.Entity(key=resp.entity.key), version=resp.version
                    )
                    for resp in resp_data
                ]
            else:
                result_type = types.EntityResult.ResultType.PROJECTION
                entity_results = [
                    types.EntityResult(
                        entity=types.Entity(
                            key=resp.entity.key,
                            properties={
                                k: v
                                for k, v in resp.entity.properties.items()
                                if k in projection_fields
                            },
                        ),
                        version=resp.version,
                    )
                    for resp in resp_data
                ]
        else:
            result_type = types.EntityResult.ResultType.FULL
            entity_results = [
                types.EntityResult(entity=resp.entity, version=resp.version)
                for resp in resp_data
            ]

        return types.RunQueryResponse(
            batch=types.QueryResultBatch(
                entity_result_type=result_type,
                entity_results=entity_results,
                snapshot_version=self.store.seqid(transaction_id),
            )
        )

    def _matches_filter(
        self, stored_obj: _StoredObject, query_filter: types.Filter
    ) -> bool:
        filter_type = query_filter.WhichOneof("filter_type")
        if filter_type is None:
            # If doing a query for all entities, filter will be None
            return True

        assert filter_type in ["property_filter", "composite_filter"]
        if filter_type == "property_filter":
            return self._matches_property_filter(
                stored_obj, query_filter.property_filter
            )
        elif filter_type == "composite_filter":
            return self._matches_composite_filter(
                stored_obj, query_filter.composite_filter
            )
        return False

    def _matches_composite_filter(
        self, stored_obj: _StoredObject, comp_filter: types.CompositeFilter
    ) -> bool:
        assert comp_filter.op == types.CompositeFilter.Operator.AND
        results = [self._matches_filter(stored_obj, f) for f in comp_filter.filters]
        return all(results)

    def _matches_property_filter(
        self, stored_obj: _StoredObject, prop_filter: types.PropertyFilter
    ) -> bool:
        for prop_name, prop_val_pb in stored_obj.entity.properties.items():
            prop_val = ds_helpers._get_value_from_value_pb(prop_val_pb)
            if prop_name == prop_filter.property.name:
                op = prop_filter.op
                filter_val = ds_helpers._get_value_from_value_pb(prop_filter.value)
                method_name = self._OPERATOR_TO_CMP_METHOD_NAME.get(op)
                assert method_name
                return getattr(prop_val, method_name)(filter_val)
        return False



import copy
import uuid
from google.cloud.datastore_v1 import types
from typing import Dict, Iterable, Optional, Tuple, List
from ._stored_object import _StoredObject
from ._transactions import _InFlightTransaction, _TransactionType


class _InMemoryStore(object):

    _seqid: int
    _store: Dict[str, _StoredObject]
    _transactions: Dict[bytes, _InFlightTransaction]

    def __init__(self) -> None:
        self._seqid = 0
        self._store = {}
        self._transactions = {}

    def seqid(self, transaction_id: Optional[bytes]) -> int:
        if transaction_id and transaction_id in self._transactions:
            return self._transactions[transaction_id].initial_seqid
        return self._seqid

    def put(
        self,
        ds_entity: types.Entity,
        entity_version: int,
        transaction_id: Optional[bytes],
    ) -> None:
        key_str = ds_entity.key.SerializeToString()
        if transaction_id and transaction_id in self._transactions:
            mutation = types.Mutation(upsert=ds_entity,)
            self._transactions[transaction_id].mutations.append(mutation)
        else:
            self._seqid += 1
            self._store[key_str] = _StoredObject(
                entity=ds_entity, version=entity_version
            )

    def get(
        self, key: types.Key, transaction_id: Optional[bytes]
    ) -> Optional[_StoredObject]:
        key_str = key.SerializeToString()
        if transaction_id and transaction_id in self._transactions:
            return self._transactions[transaction_id].snapshot.get(key_str)
        else:
            return self._store.get(key_str)

    def delete(self, key: types.Key, transaction_id: Optional[bytes]) -> None:
        key_str = key.SerializeToString()
        if transaction_id and transaction_id in self._transactions:
            mutation = types.Mutation(delete=key,)
            self._transactions[transaction_id].mutations.append(mutation)
        else:
            if key_str in self._store:
                del self._store[key_str]

    def items(self, transaction_id: bytes) -> Iterable[Tuple[str, _StoredObject]]:
        if transaction_id in self._transactions:
            return self._transactions[transaction_id].snapshot.items()
        else:
            return self._store.items()

    def beginTransaction(self, mode: _TransactionType) -> bytes:
        transaction_id = uuid.uuid1().bytes
        self._transactions[transaction_id] = _InFlightTransaction(
            mode=mode,
            initial_seqid=self._seqid,
            snapshot=copy.deepcopy(self._store),
            mutations=[],
        )
        return transaction_id

    def commitTransaction(
        self, transaction_id: bytes, final_mutations: List[types.Mutation]
    ) -> List[types.MutationResult]:
        if transaction_id != b"" and transaction_id in self._transactions:
            transaction = self._transactions[transaction_id]

            # Apply OCC (make sure no entities touched in the transaction have been modified since it began)
            # For now, just assume that no writes at all are allowed
            assert self._seqid == transaction.initial_seqid

            for mutation in transaction.mutations:
                self._applyMutation(mutation)

            del self._transactions[transaction_id]

        return [self._applyMutation(m) for m in final_mutations]

    def rollbackTransaction(self, transaction_id: bytes) -> None:
        assert transaction_id in self._transactions
        del self._transactions[transaction_id]

    def _applyMutation(self, mutation: types.Mutation) -> types.MutationResult:
        # TODO will need to potentially do key assignment for insert/upsert
        mutation_key = self._mutation_key(mutation)
        existing_data = self.get(mutation_key, None)
        operation = mutation.WhichOneof("operation")
        if operation == "insert":
            # TODO figure out how to properly express error handling
            assert existing_data is None
            self.put(mutation.insert, 0, None)
            return types.MutationResult(key=mutation_key, version=0)
        elif operation == "update":
            # TODO figure out better error handling
            assert existing_data is not None
            if existing_data.version != mutation.base_version:
                return self._mutation_conflict(mutation_key, existing_data.version)
            new_version = existing_data.version + 1
            self.put(mutation.upsert, new_version, None)
            return types.MutationResult(key=mutation_key, version=new_version)
        elif operation == "upsert":
            if existing_data and existing_data.version != mutation.base_version:
                return self._mutation_conflict(mutation_key, existing_data.version)
            new_version = existing_data.version + 1 if existing_data else 0
            self.put(mutation.upsert, new_version, None)
            return types.MutationResult(key=mutation_key, version=new_version)
        elif operation == "delete":
            new_version = existing_data.version + 1 if existing_data else 0
            self.delete(mutation_key, None)
            return types.MutationResult(key=mutation_key, version=new_version)

    def _mutation_key(self, mutation: types.Mutation) -> types.Key:
        operation = mutation.WhichOneof("operation")
        if operation == "insert":
            return mutation.insert.key
        elif operation == "update":
            return mutation.update.key
        elif operation == "upsert":
            return mutation.upsert.key
        elif operation == "delete":
            return mutation.delete.key

    def _mutation_conflict(
        self, key: types.Key, old_version: int
    ) -> types.MutationResult:
        return types.MutationResult(
            key=key, version=old_version, conflict_detected=True,
        )



import enum
from google.cloud.datastore_v1 import types
from typing import Dict, List, NamedTuple
from ._stored_object import _StoredObject


class _TransactionType(enum.Enum):
    READ_ONLY = 0
    READ_WRITE = 1


class _InFlightTransaction(NamedTuple):
    mode: _TransactionType
    initial_seqid: int
    snapshot: Dict[str, _StoredObject]
    mutations: List[types.Mutation]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="InMemoryCloudDatastoreStub",
    version="0.0.5",
    description="An in-memory stub implementation of Google Cloud Datastore for use in unit tests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/phil-lopreiato/google-cloud-datastore-stub/",
    packages=setuptools.find_packages(),
    python_requires=">=3",
    install_requires=["google-cloud-ndb == 1.2.1"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "black", "pyre-check", "flake8", "mypy"],
)



import grpc


class InstantFuture(grpc.Future):
    def __init__(self, resp):
        self.resp = resp

    def cancel(self):
        return False

    def cancelled(self):
        return False

    def running(self):
        return False

    def done(self):
        return True

    def result(self, timeout=None):
        return self.resp

    def exception(self, timeout=None):
        return None

    def traceback(self, timeout=None):
        return None

    def add_done_callback(self, fn):
        fn(self)
